unit ServerLists;

interface

uses
  ComCtrls, Classes, ExtCtrls, IdHttp;

type

  TServerList = class
  private
    FListView: TListView;
    FList: TStringList;
    FTimer: TTimer;
    itr: integer;
    procedure TimerProc(Sender: TObject);
    function Ping(Address: string): string;
  protected
    procedure Refresh; virtual;
  public
    constructor Create(AListView: TListView);
    procedure StartPinging;
    procedure TerminatePinging;
    procedure Load; virtual;
    procedure Save; virtual;
    destructor Destroy; override;
  end;

  TSavedServerList = class(TServerList)
  public
    procedure AddServer(Address, Name: string); virtual;
    procedure DeleteServer(AIndex: integer); virtual;
    procedure DeleteAll; virtual;
  end;

  TFavoriteServerList = class(TSavedServerList)
  public
    procedure Load; override;
    procedure Save; override;
    procedure AddServer(Address, Name: string); override;
    procedure DeleteServer(AIndex: integer); override;
  end;

  TVisitServerList = class(TSavedServerList)
  public
    procedure Load; override;
    procedure Save; override;
    procedure AddServer(Address, Name: string); override;
  end;

  TStatusProc = procedure(const Result: string) of object;

  TDownloadingServerList = class(TServerList)
  private
    FColumnToSort: integer;
    FAssending: boolean;
    FHttp: TIdHttp;
    FStatusProc: TStatusProc;
    procedure CompareProc(Sender: TObject; Item1, Item2: TListItem; Data: Integer; var Compare: Integer);
    procedure ColumnClick(Sender: TObject; Column: TListColumn);
    function DownloadList: string;
    procedure WriteStatus(const AResultStr: string);
  protected
    function ListUrl: string; virtual; abstract;
    procedure Refresh; override;
  public
    constructor Create(AListView: TListView);
    procedure SetStatusProc(AStatusProc: TStatusProc);
    procedure Load; override;
    destructor Destroy; override;
  end;

  TOfficialServerList = class(TDownloadingServerList)
  protected
    function ListUrl: string; override;
  end;

  TUnofficialServerList = class(TDownloadingServerList)
  protected
    function ListUrl: string; override;
  end;

implementation

uses uCore, sysutils, usettings, commonutils, strutils, uintfcommon,
  idstack, log;

{ TServerList }

constructor TServerList.Create(AListView: TListView);
begin
  FListView := AListView;
  FList := TStringList.Create;
  FList.NameValueSeparator := #30;
  FTimer := TTimer.Create(nil);
  FTimer.Enabled := false;
  FTimer.Interval := 100;
  FTimer.OnTimer := TimerProc;
end;

destructor TServerList.Destroy;
begin
  FTimer.Free;
  FList.Free;
  inherited;
end;

procedure TServerList.Load;
begin
end;

function TServerList.Ping(Address: string): string;
var
  rslt: integer;
begin
  rslt := core.CheckPing(address);
  if rslt = -1 then result := 'x' else result := inttostr(rslt);
end;

procedure TServerList.Refresh;
var
  i, j: integer;
  svrls: tstrings;
begin
  flistview.Clear;
  for I := 0 to flist.Count - 1 do
  begin
    svrls := DivValue(flist.ValueFromIndex[i]);
    try
      with FListView.Items.Add do
      begin
        Caption := flist.Names[i];
        for j := 0 to svrls.Count - 1 do
          SubItems.Add(svrls.Strings[j]);
        SubItems.Add('?');
      end;
    finally
      svrls.Free;
    end;
  end;
end;

procedure TServerList.Save;
begin

end;

procedure TServerList.startPinging;
begin
  itr := 0;
  while itr < flistview.Items.Count do
  begin
    if flistview.Items.Item[itr].SubItems[flistview.Items.Item[itr].SubItems.Count-1] = '?' then break;
    inc(itr);
  end;
  FTimer.Enabled := true;
end;

procedure TServerList.TerminatePinging;
begin
  FTimer.Enabled := false;
end;

procedure TServerList.TimerProc(Sender: TObject);
var
  address: string;
begin
  if itr >= FListView.Items.Count then
  begin
    FTimer.Enabled := false;
    exit;
  end;
  address := FListView.Items.Item[itr].SubItems[0];
  FListView.Items.Item[itr].SubItems[FListView.Items.Item[itr].SubItems.Count-1] := ping(address);
  inc(itr);
end;

{ TFavoriteServerList }

procedure TFavoriteServerList.AddServer(Address, Name: string);
begin
  FList.Add(Name+#30+Address);
  Refresh
end;

procedure TFavoriteServerList.DeleteServer(AIndex: integer);
begin
  FList.Delete(AIndex);
  Refresh;
end;

procedure TFavoriteServerList.Load;
begin
  Settings.GetFavServerList(FList);
  Refresh;
  inherited;
end;

procedure TFavoriteServerList.Save;
begin
  inherited;
  Settings.SetFavServerList(FList);
end;

{ TVisitServerList }

procedure TVisitServerList.AddServer(Address, Name: string);
var
  I: Integer;
  found: boolean;
begin
  {record visit}
  found := false;
  for I := 0 to FListView.Items.Count - 1 do
  begin
    if flistview.Items.Item[i].SubItems[0] = Address then
    begin
      found := true;
      with DivValue(FList.ValueFromIndex[i]) do
      begin
        Strings[1] := inttostr(strtoint(Strings[1])+1);
        FList.ValueFromIndex[i] := delimitedText;
        FList.Exchange(0, i);
        Free;
        break;
      end;
    end;
  end;
  if not found then FList.Insert(0, Name+#30+Address+';1');
  Refresh
end;

procedure TVisitServerList.Load;
begin
  Settings.GetVisitServerList(FList);
  Refresh;
  inherited;
end;

procedure TVisitServerList.Save;
begin
  inherited;
  Settings.SetVisitServerList(FList);
end;

{ TCurrentServerList }

procedure TDownloadingServerList.ColumnClick(Sender: TObject; Column: TListColumn);
begin
  FAssending := not FAssending;
  FColumnToSort := Column.Index;
  FListView.AlphaSort;
end;

procedure TDownloadingServerList.CompareProc(Sender: TObject; Item1,
  Item2: TListItem; Data: Integer; var Compare: Integer);
var locationtexts: array of string;
function IsInnerLocation(AText: string): boolean;
var
  i: integer;
begin
  result := false;
  for i:= 0 to length(locationtexts)-1 do
  begin
    if ansicontainstext(atext, locationtexts[i]) then
    begin
      result := true;
      exit;
    end;
  end;
end;
var
  ix: integer;
  s1, s2: string;
  n, m: integer;
  i: integer;
begin
  with divvalue(intfcommon.LocationTexts) do
  begin
    setlength(locationtexts, count);
    for i:= 0 to count-1 do
      locationtexts[i] := strings[i];
    free;
  end;
  {񱳹ڿ }
  case fcolumntosort of
    0: begin s1 := Item1.Caption; s2 := Item2.Caption; end;
    else
      begin
        if fcolumntosort = -1 then ix := 4
        else ix := FColumnToSort - 1;
        if (ix >= item1.SubItems.Count) or (ix >= item2.SubItems.Count) then exit;
        s1 := Item1.SubItems[ix]; s2 := Item2.SubItems[ix];
      end;
  end;
  {Į  }
  case FColumnToSort of
    -1:
    begin
      if isinnerlocation(s1) then
      begin
        n := maxint;
        item1.Data := pointer(17);
      end
      else n := StrToIntDef(LeftStr(item1.SubItems[1], Pos('/', item1.SubItems[1])-1), 0);
      if isinnerlocation(s2) then
      begin
         m := maxint;
         item2.Data := pointer(17);
      end
      else m := StrToIntDef(LeftStr(item2.SubItems[1], Pos('/', item2.SubItems[1])-1), 0);
      if n > m then compare := 1 else if n = m then compare := 0 else compare := -1;
    end;
    2:
    begin
      n := StrToIntDef(LeftStr(s1, Pos('/', s1)-1), 0);
      m := StrToIntDef(LeftStr(s2, Pos('/', s2)-1), 0);
      if n > m then compare := 1 else if n = m then compare := 0 else compare := -1;
    end;
    3, 6:
    begin
      n := StrToIntDef(s1, 1001);
      m := StrToIntDef(s2, 1001);
      if n > m then compare := 1 else if n = m then compare := 0 else compare := -1;
    end;
    else
    begin
      Compare := CompareText(s1, s2);
    end;
  end;
  { }
  if not FAssending then Compare := Compare * -1;
end;

constructor TDownloadingServerList.Create(AListView: TListView);
begin
  inherited Create(AListView);
  AListView.OnCompare := CompareProc;
  AListView.OnColumnClick := ColumnClick;
  FAssending := false;
  fhttp := tidhttp.Create(nil);
end;

destructor TDownloadingServerList.Destroy;
begin
  fhttp.Free;
  inherited;
end;

function TDownloadingServerList.DownloadList: string;
var
  ss:tstringstream;
begin
  result := '';
  ss:=tstringstream.Create('');
  try
    try
      fHTTP.Get(listurl,ss);
      result := ss.DataString;
    except
      on E: EIdSocketError do writeln(E.Message);
    end;
  finally
    ss.Free;
  end;
end;

procedure TDownloadingServerList.Load;
var
  sl: tstringlist;
  s: string;
  i: integer;
begin
  FAssending := false;
  WriteStatus(' ޴ ...');
  { ٿ }
  sl := tstringlist.Create;
  sl.Text := DownloadList;
  if sl.Text <> '' then
  begin
    writestatus('м...'); sleep(500);
    FList.Clear;
    try
      // Ľ   ..
      i:=0;
      if sl[i] = '|||||||' then inc(i);
      while i < sl.Count-1 do
      begin
        s := sl[i]; inc(i);
        if not ansicontainsstr(s, ';') then
        begin
          s := s + #30 + sl[i]; inc(i);
          FList.Add(s);
        end;
      end;
      WriteStatus('߰...'); sleep(500);
      Refresh;
      writestatus(inttostr(FList.Count)+'   ã.');
    except
      WriteStatus(' Ľ .');
    end;
  end
  else WriteStatus(' ޱ .');
  sl.Free;
end;

procedure TDownloadingServerList.Refresh;
begin
  inherited;
  writestatus('...');
  FColumnToSort := -1;
  FListView.AlphaSort;
end;

procedure TDownloadingServerList.SetStatusProc(AStatusProc: TStatusProc);
begin
  FStatusProc := AStatusProc;
end;

procedure TDownloadingServerList.WriteStatus(const AResultStr: string);
begin
  if assigned(fStatusProc) then fStatusProc(aresultstr);
end;

{ TOfficialServerList }

function TOfficialServerList.ListUrl: string;
begin
  result := 'http://kaillera.com/raw_server_list2.php';
end;

{ TUnofficialServerList }

function TUnofficialServerList.ListUrl: string;
begin
  result := 'http://master.anti3d.com/raw_server_list2.php';
end;

{ TSavedServerList }

procedure TSavedServerList.AddServer(Address, Name: string);
begin

end;

procedure TSavedServerList.DeleteAll;
begin
  flist.Clear;
  refresh;
end;

procedure TSavedServerList.DeleteServer(AIndex: integer);
begin

end;

end.
