unit Containers;

interface

uses Types, SysUtils;

type

  EContainerOutOfData = class(Exception)
  public
    constructor Create(const Msg: string);
  end;

  TBaseContainer = object
  protected
    FData: PByteArray;
    FSize: Cardinal;
    FCapacity: Cardinal;
    procedure GetinData(const Buf; ASize: integer); virtual;
    procedure GetoutData(var Buf; ASize: integer); virtual;
  public
    constructor Init;
    destructor Done; virtual;
    procedure Push(Value: Byte); overload;
    procedure Push(Value: Word); overload;
    procedure Push(Value: LongWord); overload;
    procedure Push(Value: TByteDynArray); overload;
    procedure Push(const Value; ASize: integer); overload;
    procedure Pop(var Value: Byte); overload;
    procedure Pop(var Value: Word); overload;
    procedure Pop(var Value: LongWord); overload;
    procedure Pop(var Value: TByteDynArray); overload;
    procedure Pop(var Value; ASize: integer); overload;
    function GetBinaryData: TByteDynArray; virtual; abstract;
    procedure SetBinaryData(const APackedData; ASize: integer); virtual; abstract;
    function Size: integer; virtual;
    procedure Clear; virtual;
  end;

  PByteBuffer = ^TByteBuffer;
  TByteBuffer = object(TBaseContainer)
  private
    FHead: cardinal;
  protected
    procedure GetinData(const Buf; ASize: integer); virtual;
    procedure GetoutData(var Buf; ASize: integer); virtual;
  public
    constructor Init;
    function GetBinaryData: TByteDynArray; virtual;
    procedure SetBinaryData(const APackedData; ASize: integer); virtual;
    procedure Push(const Value: ShortString); overload;
    function PopString: ShortString;
    function Size: integer; virtual;
    procedure Clear; virtual;
  end;

  TVarStack = object(TBaseContainer)
  protected
    procedure GetinData(const Buf; ASize: integer); virtual;
    procedure GetoutData(var Buf; ASize: integer); virtual;
  public
    function GetBinaryData: TByteDynArray; virtual;
    procedure SetBinaryData(const APackedData; ASize: integer); virtual;
  end;

  TPriorityItem = record
    Priority: integer;
    Item: pointer;
  end;
  PPriorityItems = ^TPriorityItems;
  TPriorityItems = array[1..MaxInt div 16] of TPriorityItem;

  TPriorityQueue = class
  private
    FItems: PPriorityItems;
    FIndex: integer;
    procedure Swap(Index1, Index2: integer);
  public
    constructor Create(AMaxCount: integer);
    destructor Destroy; override;
    procedure Push(AItem: pointer; Priority: integer);
    function Pop: pointer;
    property Count: integer read findex;
    procedure Clear;
  end;

  TDataBuffer = class
  private
    FHead: Byte;
    FTail: Byte;
    FData: array[0..$FF] of ansichar;
  public
    constructor Create;
    procedure Push(AInData: pansichar; ASize: integer);
    function Pop(AOutData: pansichar; ASize: integer): integer;
    function Size: integer;
    procedure Clear;
  end;

implementation

const GrowSize = $FF;

procedure MoveRev(const Source; var Dest; Count: integer);
var
  s, d: PAnsiChar;
  i, j: integer;
begin
  s := PAnsiChar(@Source);
  d := PAnsiChar(@Dest);
  j:=0;
  for I := Count-1 downto 0 do
  begin
    d[j] := s[i];
    inc(j);
  end;
end;

{ TVector }

procedure TBaseContainer.Push(Value: TByteDynArray);
begin
  if length(Value) <> 0 then
    GetinData(Value[0], length(Value));
end;

procedure TBaseContainer.Push(Value: Byte);
begin
  GetinData(Value, sizeof(Value));
end;

procedure TBaseContainer.Push(Value: Word);
begin
  GetinData(Value, sizeof(Value));
end;

procedure TBaseContainer.Push(Value: LongWord);
begin
  GetinData(Value, sizeof(Value));
end;

procedure TBaseContainer.Pop(var Value: Byte);
begin
  Value := 0;
  GetoutData(Value, 1);
end;

procedure TBaseContainer.Pop(var Value: TByteDynArray);
begin
  GetoutData(Value[0], length(Value));
end;

procedure TBaseContainer.Clear;
begin
  fsize := 0;
end;

constructor TBaseContainer.Init;
begin
  FData := AllocMem(GrowSize);
  FCapacity := GrowSize;
  FSize := 0;
end;

destructor TBaseContainer.Done;
begin
  FreeMem(FData);
end;

procedure TBaseContainer.GetinData(const Buf; ASize: integer);
begin
  while (fcapacity - FSize) < ASize do
  begin
    inc(fcapacity, growsize);
    ReallocMem(FData, fcapacity);
  end;
end;

procedure TBaseContainer.GetoutData(var Buf; ASize: integer);
begin
  if Size < ASize then
    raise EContainerOutOfData.Create('Required Size of Data is not Exists in Container.');
end;

function TBaseContainer.Size: integer;
begin
  result := FSize;
end;

procedure TBaseContainer.Pop(var Value: Word);
begin
  Value := 0;
  GetoutData(Value, 2);
end;

procedure TBaseContainer.Pop(var Value; ASize: integer);
begin
  GetoutData(Value, ASize);
end;

procedure TBaseContainer.Pop(var Value: LongWord);
begin
  GetoutData(Value, 4);
end;

procedure TBaseContainer.Push(const Value; ASize: integer);
begin
  GetinData(Value, ASize);
end;

{ TByteBuffer }

procedure TByteBuffer.GetinData(const Buf; ASize: integer);
begin
  inherited;
  Move(Buf, FData[FSize], ASize);
  FSize := FSize + ASize;
end;

procedure TByteBuffer.GetoutData(var Buf; ASize: integer);
begin
  inherited;
  move(FData[FHead], Buf, ASize);
  inc(FHead, ASize);
end;

function TByteBuffer.GetBinaryData: TByteDynArray;
begin
  setlength(result, Size); //  빰  ƴ϶ ӹ ü ŭ Ҵ
  move(fdata[fhead], result[0], size);
end;

function TByteBuffer.PopString: ShortString;
begin
  result := PAnsiChar(@FData[FHead]);
  inc(FHead, strlen(PAnsiChar(@FData[FHead]))+1);
end;

procedure TByteBuffer.Push(const Value: ShortString);
const zero: byte = $00;
begin
  if Value <> '' then GetinData(Value[1], length(Value));
  GetinData(zero, 1);
end;

procedure TByteBuffer.SetBinaryData(const APackedData; ASize: integer);
begin
  fsize := 0;
  FHead := 0;
  getindata(apackeddata, asize);
end;

function TByteBuffer.Size: integer;
begin
  result := fsize - fhead;
end;

constructor TByteBuffer.Init;
begin
  inherited Init;
  fhead := 0;
end;

procedure TByteBuffer.Clear;
begin
  inherited clear;
  fhead := 0;
end;

{ TVarStack }

procedure TVarStack.GetinData(const Buf; ASize: integer);
begin
  inherited;
  MoveRev(Buf, FData[FSize], ASize);
  FSize := FSize + ASize;
end;

procedure TVarStack.GetoutData(var Buf; ASize: integer);
begin
  inherited;
  MoveRev(FData[FSize-ASize], Buf, ASize);
  FSize := FSize - ASize;
end;

function TVarStack.GetBinaryData: TByteDynArray;
begin
  inherited GetBinaryData;
  setlength(result, FSize);
  MoveRev(FData[0], result[0], FSize);
end;

procedure TVarStack.SetBinaryData(const APackedData; ASize: integer);
begin
//  FSize := asize;
//  setlength(FData, FSize);
//  MoveRev(APackedData[0], FData[0], FSize);
end;

{ TPriorityQueue }

procedure TPriorityQueue.Clear;
begin
  findex := 0;
end;

constructor TPriorityQueue.Create(AMaxCount: integer);
begin
  fitems := allocmem(amaxcount);
  clear;
end;

destructor TPriorityQueue.Destroy;
begin
  freemem(fitems);
  inherited;
end;

function TPriorityQueue.Pop: pointer;
var child_idx, parent_idx: integer;
begin
  if findex = 0 then begin result := nil; exit; end;
  result := fitems^[1].Item;
  fitems[1] := fitems[findex];
  dec(findex);
  parent_idx := 1;
  child_idx := parent_idx * 2;
  while (child_idx < findex) and (fitems^[parent_idx].Priority > fitems^[child_idx].Priority) do
  begin
    swap(parent_idx, child_idx);
    parent_idx := child_idx * 2;
    child_idx := parent_idx * 2;
  end;
end;

procedure TPriorityQueue.Push(AItem: pointer; Priority: integer);
var child_idx, parent_idx: integer;
begin
  inc(findex);
  fitems^[findex].Priority := priority;
  fitems^[findex].Item := aitem;
  child_idx := findex;
  parent_idx := child_idx div 2;
  while (child_idx > 1 ) and (fitems^[child_idx].Priority = fitems^[parent_idx].Priority) do
  begin
    swap(child_idx, parent_idx);
    child_idx := parent_idx;
    parent_idx := child_idx div 2;
  end;
end;

procedure TPriorityQueue.Swap(Index1, Index2: integer);
var
  tmp: tpriorityitem;
begin
  tmp := fitems^[index1];
  fitems^[index1] := fitems^[index2];
  fitems^[index2] := tmp;
end;

{ TDataBuffer }

procedure TDataBuffer.Clear;
begin
  ftail := fhead;
end;

constructor TDataBuffer.Create;
begin
  fhead := 0;
  ftail := 0;
end;

function TDataBuffer.Pop(AOutData: pansichar; ASize: integer): integer;
var i: integer;
begin
  if byte(ftail-fhead) < asize then begin result := 0; exit; end;
  i:=0;
  while i < asize do
  begin
    aoutdata[i] := fdata[fhead];
    inc(fhead);
    inc(i);
  end;
  result := asize;
end;

procedure TDataBuffer.Push(AInData: pansichar; ASize: integer);
var i: integer;
begin
  i:=0;
  while i < asize do
  begin
    fdata[ftail] := aIndata[i];
    inc(ftail);
    inc(i);
  end;
end;

function TDataBuffer.Size: integer;
begin
  result := byte(ftail-fhead);
end;

{ EContainerOutOfData }

constructor EContainerOutOfData.Create(const Msg: string);
begin
  inherited create(msg);
end;

end.
