unit Users;

interface

uses
  classes, sysutils, idudpserver, idglobal, idsockethandle,
  kaillerautils, instructions2, containers, commonutils, games, timer;

type

  { TUserManager }

  TUser = class;
  TEachUserProc = reference to procedure(AUser: TUser);
  TEachUser = procedure(AEachUserProc: TEachUserProc) of object;

  TUserManager = class
  private
    UserList: TThreadList;
    ZombiKiller: TTimer;
    GameIDPool: TNumberPool;
    UserIDPool: TNumberPool;
    procedure KillZombi(Sender: TObject);
  public
    constructor Create();
    destructor Destroy; override;
    function AddUser: TUser;
    procedure EachUser(AEachUserProc: TEachUserProc);
    function GetGameID: cardinal;
  end;

  { TUser }

  TUser = class
  private
    { Connection Info }
    FUserIP: string;
    FUserPort: TPort;
    FAllowedPort: TPort;
    { objects }
    UDPServer: TIdUDPServer;
    FIncomingInstructions: TInstructionRecvQueue;
    FOutgoingInstructions: TInstructionSendQueue;
    FGameDataSendCache: TGameDataCache;
    FGameDataRecvCache: TGameDataCache;
    FTimeBand: integer;
    PingChecker: TDelayChecker;
    FPlayerNo: integer;
    FEachUser: TEachUser;
    procedure ProcessInstruction(AThread: TIdUDPListenerThread; AData: TIdBytes; ABinding: TIdSocketHandle);
    procedure SendInstructionToAll(AInstruction: PInstruction);
  public
    { User Info }
    UserID: integer;
    UserName: string;
    AppName: string;
    ConnectionType: Byte;
    Ping: integer;
    Status: integer;
    //
    UserAlive: boolean;
    Game: TGame;
    constructor Create(AUserID: cardinal);
    destructor Destroy; override;
    procedure Activate(AUserIP: string; AllowedPort: TPort);
    procedure SendInstruction(AInstruction: PInstruction; FreeIns: boolean = true);
    procedure SetEachUser(AEachUser: TEachUser);
  end;

implementation

uses log;

{ TUserList }

constructor TUserManager.Create();
begin
  GameIDPool := tnumberpool.Create(1, 100);
  useridpool := tnumberpool.Create(1, 300);
  UserList := tthreadlist.Create;
  ZombiKiller := ttimer.Create();
  ZombiKiller.OnTimer := killzombi;
  ZombiKiller.Interval := 1000;
  ZombiKiller.Enabled := true;
end;

destructor TUserManager.Destroy;
var i: integer; ls: tlist;
begin
  ZombiKiller.Free;
  useridpool.Free;
  GameIDPool.Free;
  ls := userlist.LockList;
  for i:= 0 to ls.Count-1 do
    tobject(ls.Items[i]).Free;
  UserList.free;
  inherited Destroy;
end;

procedure TUserManager.EachUser(AEachUserProc: TEachUserProc);
var
  i: integer;
  ls: tlist;
begin
  ls := userlist.LockList;
  for i:= 0 to ls.Count-1 do
    aeachuserproc(tuser(ls.Items[i]));
  userlist.UnlockList;
end;

function TUserManager.GetGameID: cardinal;
begin
  result := gameidpool.PullNumber;
end;

procedure TUserManager.KillZombi(Sender: TObject);
var
  i: Integer;
  ls: tlist;
begin
  ls := userlist.LockList;
  i:=0;
  while i < ls.Count do
  begin
    with tuser(ls.Items[i]) do
      if not UserAlive then
      begin
        writeln(format('user %s is dead, deleting..',[UserName]));
        useridpool.RestoreNumber(userid);
        ls.Delete(i);
        free;
      end;
    inc(i);
  end;
  userlist.UnlockList;
end;

function TUserManager.AddUser: TUser;
begin
  result := tuser.Create(useridpool.PullNumber);
  userlist.Add(result);
  result.SetEachUser(EachUser);
end;

{ TUser }

procedure TUser.Activate(AUserIP: string; AllowedPort: TPort);
begin
  fuserip := auserip;
  fallowedport := allowedport;
  udpserver.DefaultPort := allowedport;
  udpserver.Active := true;
end;

constructor TUser.Create(AUserID: cardinal);
begin
  userid := auserid;
  fincominginstructions := tinstructionrecvqueue.create;
  foutgoinginstructions := tinstructionsendqueue.create;
  useralive := true;
  fgamedatasendcache := tgamedatacache.Create;
  fgamedatarecvcache := tgamedatacache.Create;
  pingchecker := TDelayChecker.Create(4);
  udpserver := tidudpserver.Create(nil);
  udpserver.OnUDPRead := ProcessInstruction;
  udpserver.ThreadedEvent := true;
end;

destructor TUser.Destroy;
begin
  udpserver.free;
  pingchecker.Free;
  foutgoinginstructions.free;
  fincominginstructions.free;
  fgamedatasendcache.Free;
  fgamedatarecvcache.Free;
  inherited Destroy;
end;

procedure TUser.SendInstruction(AInstruction: PInstruction; FreeIns: boolean);
var
  pkt: tpacket;
begin
  foutgoinginstructions.Push(ainstruction);
  pkt := foutgoinginstructions.GetPacket;
  writeln(format('send %s to %s:%d',[ainstruction.GetIDString, fuserip, fuserport]));
  udpserver.SendBuffer(fuserip, fuserport, tidbytes(pkt));
  if freeins then dispose(ainstruction, done);
end;

procedure TUser.SendInstructionToAll(AInstruction: PInstruction);
begin
  FEachUser(
    procedure(AUser: TUser)
    begin
      auser.SendInstruction(ainstruction);
    end
  );
end;

procedure TUser.SetEachUser(AEachUser: TEachUser);
begin
  feachuser := aeachuser;
end;

procedure TUser.ProcessInstruction(AThread: TIdUDPListenerThread; AData: TIdBytes;
  ABinding: TIdSocketHandle);
  function MakeULS: ins_uls;
  var ins: ins_uls;
  begin
    {  Ӹϰ    - ĳø  }
    ins := new(ins_uls,init);
    fEachUser(
      procedure(auser: tuser)
      begin
        if auser.UserID <> UserID then // ڽ 
          with auser do ins^.AddUser(UserName, Ping, Status, UserID, ConnectionType);
      end
    );
    result := ins;
  end;
var
  i: integer;
  ins: pinstruction;
begin
  { recv instruction }
  fincominginstructions.ExtractData(@adata[0], length(adata));
  ins := fincominginstructions.Pop;
  { dispatch instruction }
  if ins <> nil then
  begin
    writeln('disp: ', ins.getidstring);
    case ins.getid of
      lrq:
      begin
        fuserport := abinding.PeerPort;
        with ins_lrq(ins)^ do
        begin
          self.UserName:=UserName;
          self.Connectiontype:=Connection;
          self.appname := AppName;
        end;
        sendinstruction(new(ins_ping, init));
        pingchecker.CheckIn;
      end;
      pong:
      begin
        pingchecker.CheckOut;
        if pingchecker.Count < 4 then begin
          sendinstruction(new(ins_ping, init));
          pingchecker.CheckIn;
        end else begin
          { Login Success! }
          Ping:= pingchecker.GetPing;
          fTimeBand := trunc(0.06 * Ping / Connectiontype) + 1;
          { send uls }
          sendinstruction(makeuls);
          { send ujn myself }
          sendinstructiontoall(new(ins_ujn,Init(UserName, UserID, Ping, Connectiontype)));
        end;
      end;
      ULV: with ins_ulv(ins)^ do
      begin
        sendinstructiontoall(new(ins_ulv,Init(LeaveMessage, UserName, UserID)));
        useralive := false;
      end;
      TMO: ;
      CHT:
      begin
        Sendinstructiontoall(
          new(ins_cht,Init(ins_cht(ins).Message, UserName))
        );
      end;
      GCR:
      begin
        game := tgame.Create(ins_gcr(ins).GameName, 0, self.appname, self.UserName);
        fplayerno := 1;
        with game do
          sendinstructiontoall(
            new(ins_gcr,Init(Title, AppName, GameId, Owner))
          );
        SendInstruction(new(ins_pjn,init(connectiontype, 0, UserName, Ping, UserID)));
      end;
      GSA:
      begin
        fgamedatasendcache.Clear;
        fgamedatarecvcache.Clear;
        sendinstruction(new(ins_gst,Init(game.GameId, 2, fplayerno, 2)));
        sendinstruction(new(ins_gsa,init(ftimeband, fplayerno, 2)));
      end;
      RDY:
      begin
        game.GetReady(fplayerno);
        sendinstruction(new(ins_gst,Init(game.gameId, 3, 1, 2)));
        sendinstruction(new(ins_rdy,init));
      end;
      GDA:
      begin
        { recv }
        with ins_gda(ins)^ do
        begin
          game.PutData(Data, Size, fplayerno);
          i := fgamedatarecvcache.Find(data, size);
        end;
        if i = -1 then // ĳ   ߰
          fgamedatarecvcache.Push(ins)
        else
          dispose(ins, done);
        { send back}
        ins := new(ins_gda,init);
        with ins_gda(ins)^ do
        begin
          game.GetData(data^, size, 1);
          i := fgamedatasendcache.Find(data, size);
          if i = -1 then
          begin
            fgamedatasendcache.Add(ins);
            sendinstruction(ins, false);
          end
          else
          begin
            dispose(ins, done);
            sendinstruction(new(ins_cda,init(i)));
          end;
        end;
        exit;
      end;
      CDA:
      begin
        { recv }
        with fgamedatarecvcache.Get(ins_cda(ins).Index)^ do
          game.PutData(data^, size, fplayerno);
        dispose(ins,done);
        { send back }
        ins := new(ins_gda, init);
        with ins_gda(ins)^ do
        begin
          game.GetData(data^, size, 1);
          i := fgamedatasendcache.Find(data, size);
          if i = -1 then
          begin
            fgamedatasendcache.Add(ins);
            sendinstruction(ins, false);
          end
          else
          begin
            dispose(ins,done);
            sendinstruction(new(ins_cda,init(i)));
          end;
        end;
        exit;
      end;
      PLV:
      begin
        if fplayerno = 1 then
        begin
          sendinstruction(new(ins_gcl,Init(game.gameId)));
          game.Free;
        end
        else // gst  
        ;
        sendinstruction(new(ins_plv,Init(UserName, UserID)));
      end;
      PJN:
      begin
      end;
    end;
    dispose(ins, done);
  end;
end;

end.
